import { world, system, BlockPermutation, ItemStack } from "@minecraft/server";
import * as ui from "@minecraft/server-ui";

const airLikeBlocks = ['minecraft:air', 'minecraft:water', 'minecraft:flowing_water', 'minecraft:flowing_lava', 'minecraft:tall_grass', 'minecraft:short_grass', 'minecraft:seagrass'];

const hfSnowballMap = new Map();
const hfDataMap = new Map();

const blockNames = ["Auto Detect", "Acacia Log", "Acacia Planks", "Andesite", "Bedrock", "Birch Log", "Birch Planks", "Block of Coal", "Block of Diamond", "Block of Emerald", "Block of Gold", "Block of Iron", "Block of Lapis Lazuli", "Block of Redstone", "Blue Wool", "Brown Wool", "Coal Ore", "Cobblestone", "Concrete", "Coarse Dirt", "Cyan Wool", "Dark Oak Log", "Dark Oak Planks", "Diamond Block", "Diorite", "Dirt", "Emerald Block", "End Stone", "Gold Block", "Gold Ore", "Grass Block", "Gravel", "Gray Wool", "Green Wool", "Iron Block", "Iron Ore", "Jungle Log", "Jungle Planks", "Lapis Block", "Light Blue Wool", "Light Gray Wool", "Lime Wool", "Magenta Wool", "Magma Block", "Netherrack", "Oak Log", "Oak Planks", "Obsidian", "Orange Wool", "Pink Wool", "Podzol", "Polished Andesite", "Polished Diorite", "Polished Granite", "Prismarine", "Purpur Block", "Red Sand", "Red Wool", "Redstone Block", "Sand", "Sea Lantern", "Soul Sand", "Spruce Log", "Spruce Planks", "Stone", "Terracotta", "White Wool", "Yellow Wool"];
const blockTypeIds = ["DarkBlock_Gaming", "minecraft:acacia_log", "minecraft:acacia_planks", "minecraft:andesite", "minecraft:bedrock", "minecraft:birch_log", "minecraft:birch_planks", "minecraft:coal_block", "minecraft:diamond_block", "minecraft:emerald_block", "minecraft:gold_block", "minecraft:iron_block", "minecraft:lapis_block", "minecraft:redstone_block", "minecraft:blue_wool", "minecraft:brown_wool", "minecraft:coal_ore", "minecraft:cobblestone", "minecraft:concrete", "minecraft:coarse_dirt", "minecraft:cyan_wool", "minecraft:dark_oak_log", "minecraft:dark_oak_planks", "minecraft:diamond_block", "minecraft:diorite", "minecraft:dirt", "minecraft:emerald_block", "minecraft:end_stone", "minecraft:gold_block", "minecraft:gold_ore", "minecraft:grass_block", "minecraft:gravel", "minecraft:gray_wool", "minecraft:green_wool", "minecraft:iron_block", "minecraft:iron_ore", "minecraft:jungle_log", "minecraft:jungle_planks", "minecraft:lapis_block", "minecraft:light_blue_wool", "minecraft:light_gray_wool", "minecraft:lime_wool", "minecraft:magenta_wool", "minecraft:magma_block", "minecraft:netherrack", "minecraft:oak_log", "minecraft:oak_planks", "minecraft:obsidian", "minecraft:orange_wool", "minecraft:pink_wool", "minecraft:podzol", "minecraft:polished_andesite", "minecraft:polished_diorite", "minecraft:polished_granite", "minecraft:prismarine", "minecraft:purpur_block", "minecraft:red_sand", "minecraft:red_wool", "minecraft:redstone_block", "minecraft:sand", "minecraft:sea_lantern", "minecraft:soul_sand", "minecraft:spruce_log", "minecraft:spruce_planks", "minecraft:stone", "minecraft:terracotta", "minecraft:white_wool", "minecraft:yellow_wool"];

function initHfData(player) {
	hfDataMap.set(player.id, {
		hfAnim: 1,
		hfSpeed: 1,
		hfBlock: 0,
		hfNotify: false,
		location: undefined,
		blockData: undefined,
	});
}

function hfSettings(player) {
	const hfData = hfDataMap.get(player.id);
	const hfSettingsForm = new ui.ModalFormData()
		.title("§9Hole Filler v1.3.0")
		.dropdown("Hole Filling Animation", ["Instantaneous", "Wave (Default)", "One By One (Cubic)", "Expanding Wave", "Explosion Like", "Cascading Pattern", "Random",], { defaultValueIndex: hfData.hfAnim })
		.slider("Animation Time", 1, 10, { valueStep: 1, defaultValue: hfData.hfSpeed })
		.dropdown("Filling Block", blockNames, { defaultValueIndex: hfData.hfBlock })
		.toggle("Notify no. of blocks filled", { defaultValue: hfData.hfNotify });
	hfSettingsForm.show(player).then(response => {
		if (response.canceled) {
			player.sendMessage("Please click submit!");
		}
		hfData.hfAnim = response.formValues[0];
		hfData.hfSpeed = response.formValues[1];
		hfData.hfBlock = response.formValues[2];
		hfData.hfNotify = response.formValues[3];
	});
}

world.afterEvents.itemUse.subscribe((event) => {
	const player = event.source;
	if (event.itemStack?.typeId === 'minecraft:snowball' && event.itemStack.nameTag === "Hole Filler") {
		if (!hfDataMap.get(player.id)) {
			initHfData(player)
		}
		const ent = player.dimension.getEntities({
			location: player.location,
			closest: 1,
			type: "minecraft:snowball",
		});
		const ent1 = ent[0];

		if (ent1) {
			hfSnowballMap.set(ent1.id, player.id);
		}

	}
	if (event.itemStack?.typeId === 'minecraft:stick' && event.itemStack.nameTag === "Hole Filler Settings") {
		if (!hfDataMap.get(player.id)) {
			initHfData(player)
		}
		hfSettings(player);
	}
	if (event.itemStack?.typeId === 'minecraft:dirt') {
		if (!hfDataMap.get(player.id)) {
			initHfData(player)
		}
	}
});

world.afterEvents.projectileHitBlock.subscribe(event => {
	const player = event.source;
	const hfData = hfDataMap.get(player.id);
	if (event.projectile.typeId !== 'minecraft:snowball') return;
	const snowballId = event.projectile.id;

	if (!hfSnowballMap.has(snowballId)) return;
	hfSnowballMap.delete(snowballId);

	const dimension = event.dimension;
	const hitBlock = event.getBlockHit()?.block;

	if (hfData.hfBlock === 0 && hitBlock) {
		hfData.location = hitBlock.location;
		hfData.blockData = {
			typeId: hitBlock.typeId,
			states: hitBlock.permutation.getAllStates(),
		};
	} else if (hfData.hfBlock !== 0 && hitBlock) {
		hfData.location = hitBlock.location;
	}

	const radiusNorth = 6; // Negative Z
	const radiusSouth = 6; // Positive Z
	const radiusEast = 6; // Positive X
	const radiusWest = 6; // Negative X
	const heightDown = 4; // Negative Y
	let heightUp = 0; // Positive Y, dynamically set based on surface detection

	let surfaceLevel = hfData.location.y;
	let foundSurface = false;

	// Scan for surface level by checking above the hit location
	for (let y = hfData.location.y; y <= hfData.location.y + 5; y++) {
		let isAirSurface = true;

		for (let xOffset = -3; xOffset <= 3; xOffset++) {
			for (let zOffset = -3; zOffset <= 3; zOffset++) {
				const targetLocation = {
					x: hfData.location.x + xOffset,
					y: y,
					z: hfData.location.z + zOffset
				};
				const block = dimension.getBlock(targetLocation);

				// Use the array to check for air-like blocks
				if (!airLikeBlocks.includes(block.typeId)) {
					isAirSurface = false;
					break;
				}
			}
			if (!isAirSurface) break;
		}

		if (isAirSurface) {
			heightUp = (y - 1) - hfData.location.y;
			foundSurface = true;
			surfaceLevel = y - 1;
			break;
		}
	}

	if (!foundSurface) heightUp = 5;

	const airBlocksToFill = [];

	// Collect air block locations around the selected block in all six directions
	for (let yOffset = 0; yOffset <= heightUp; yOffset++) {
		for (let xOffset = -radiusWest; xOffset <= radiusEast; xOffset++) {
			for (let zOffset = -radiusNorth; zOffset <= radiusSouth; zOffset++) {
				// For each layer, calculate the positions symmetrically around the center (selected block)
				const positiveYOffset = Math.min(hfData.location.y + yOffset, 320); // Upward direction
				const negativeYOffset = Math.max(hfData.location.y - yOffset, -64); // Downward direction

				// Upward from selected block
				const targetLocationUp = {
					x: hfData.location.x + xOffset,
					y: positiveYOffset,
					z: hfData.location.z + zOffset
				};

				// Downward from selected block
				const targetLocationDown = {
					x: hfData.location.x + xOffset,
					y: negativeYOffset,
					z: hfData.location.z + zOffset
				};

				const blockUp = dimension.getBlock(targetLocationUp);
				const blockDown = dimension.getBlock(targetLocationDown);

				// Use the array to check for air-like blocks
				if (airLikeBlocks.includes(blockUp.typeId)) {
					airBlocksToFill.push(targetLocationUp);
				}
				if (airLikeBlocks.includes(blockDown.typeId)) {
					airBlocksToFill.push(targetLocationDown);
				}
			}
		}
	}

	if (hfData.hfAnim === 0) {
		// Instantly fill all blocks from the airBlocksToFill array
		let filledCount = 0;
		for (const targetLocation of airBlocksToFill) {
			const block = dimension.getBlock(targetLocation);

			// Check if the block is air-like before filling
			if (airLikeBlocks.includes(block.typeId)) {
				if (hfData.hfBlock === 0) {
					if (hfData.hfBlock === 0) {
						block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
					} else {
						block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
					}
				} else {
					block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
				}
				filledCount++;
			}
		}
		if (hfData.hfNotify === true) {
			player.sendMessage(`Filled ${filledCount} blocks.`);
		}
	}


	if (hfData.hfAnim === 1) {
		// Start filling the blocks with a smoother wave-like placement
		let filledCount = 0;
		let maxWave = Math.max(radiusNorth, radiusSouth, radiusEast, radiusWest) + heightUp + heightDown;
		let currentWave = 0;
		const interval = 1; // Adjust this for faster or slower placement

		const fillInterval = system.runInterval(() => {
			if (filledCount < airBlocksToFill.length) {
				const blocksInThisWave = [];

				// Collect blocks for the current "wave" (blocks that are at the current layer of expansion)
				for (let i = 0; i < airBlocksToFill.length; i++) {
					const blockDistance = Math.abs(airBlocksToFill[i].x - hfData.location.x) +
						Math.abs(airBlocksToFill[i].y - hfData.location.y) +
						Math.abs(airBlocksToFill[i].z - hfData.location.z);

					if (blockDistance === currentWave) {
						blocksInThisWave.push(airBlocksToFill[i]);
					}
				}

				// Place all blocks in the current "wave"
				for (const targetLocation of blocksInThisWave) {
					const block = dimension.getBlock(targetLocation);

					// Use the array to check for air-like blocks
					if (airLikeBlocks.includes(block.typeId)) {
						if (hfData.hfBlock === 0) {
							block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
						} else {
							block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
						}
						filledCount++;
					}
				}
				currentWave++;

				// Stop when all blocks are filled or when no more waves are needed
				if (filledCount >= airBlocksToFill.length || currentWave > maxWave) {
					system.clearRun(fillInterval);
					if (hfData.hfNotify === true) {
						player.sendMessage(`Filled ${filledCount} blocks.`);
					}
				}

			} else {
				system.clearRun(fillInterval);
				if (hfData.hfNotify === true) {
					player.sendMessage(`Filled ${filledCount} blocks.`);
				}
			}
		}, hfData.hfSpeed);
	}

	if (hfData.hfAnim === 2) {
		let filledCount = 0;
		let currentIndex = 0;

		// Define the interval for block placement
		const interval = 1; // Adjust for speed (in ticks)

		const fillInterval = system.runInterval(() => {
			if (currentIndex < airBlocksToFill.length) {
				const targetLocation = airBlocksToFill[currentIndex];
				const block = dimension.getBlock(targetLocation);

				// Check if the block is air-like before filling
				if (airLikeBlocks.includes(block.typeId)) {
					if (hfData.hfBlock === 0) {
						block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
					} else {
						block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
					}
					filledCount++;
				}

				currentIndex++;

			} else {
				// Stop the interval and notify the player when all blocks are filled
				system.clearRun(fillInterval);
				if (hfData.hfNotify === true) {
					player.sendMessage(`Filled ${filledCount} blocks.`);
				}
			}
		}, hfData.hfSpeed);
	}


	if (hfData.hfAnim === 3) {
		let waveDirection = 'x'; // Can be 'x' or 'z' to determine the wave direction
		let currentWave = 0; // Track the current wave level
		const waveHeight = 2; // Height of each wave (how high the blocks will be placed)
		let filledCount = 0;

		const waveInterval = system.runInterval(() => {
			if (filledCount < airBlocksToFill.length) {
				for (const blockPos of airBlocksToFill) {
					if (waveDirection === 'x') {
						if (blockPos.x === hfData.location.x + currentWave || blockPos.x === hfData.location.x - currentWave) {
							const block = dimension.getBlock(blockPos);
							if (airLikeBlocks.includes(block.typeId)) {
								if (hfData.hfBlock === 0) {
									block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
								} else {
									block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
								}
								filledCount++;
							}
						}
					} else if (waveDirection === 'z') {
						if (blockPos.z === hfData.location.z + currentWave || blockPos.z === hfData.location.z - currentWave) {
							const block = dimension.getBlock(blockPos);
							if (airLikeBlocks.includes(block.typeId)) {
								if (hfData.hfBlock === 0) {
									block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
								} else {
									block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
								}
								filledCount++;
							}
						}
					}
				}

				currentWave++;

				// Stop the process when all blocks are filled
				if (filledCount >= airBlocksToFill.length || currentWave > 10) {
					system.clearRun(waveInterval);
					if (hfData.hfNotify === true) {
						player.sendMessage(`Filled ${filledCount} blocks.`);
					}
				}

			} else {
				system.clearRun(waveInterval);
			}
		}, hfData.hfSpeed + 1); // Adjust the interval for smoother or faster waves
	}


	if (hfData.hfAnim === 4) {
		let explosionRadius = 1; // Start with a small radius and expand outward
		let filledCount = 0;

		const explosionInterval = system.runInterval(() => {
			if (filledCount < airBlocksToFill.length) {
				for (const blockPos of airBlocksToFill) {
					const distance = Math.sqrt(
						Math.pow(blockPos.x - hfData.location.x, 2) +
						Math.pow(blockPos.y - hfData.location.y, 2) +
						Math.pow(blockPos.z - hfData.location.z, 2)
					);

					if (distance <= explosionRadius) {
						const block = dimension.getBlock(blockPos);
						if (airLikeBlocks.includes(block.typeId)) {
							if (hfData.hfBlock === 0) {
								block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
							} else {
								block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
							}
							filledCount++;
						}
					}
				}

				explosionRadius++;

				if (filledCount >= airBlocksToFill.length || explosionRadius > 10) {
					system.clearRun(explosionInterval);
					if (hfData.hfNotify === true) {
						player.sendMessage(`Filled ${filledCount} blocks.`);
					}
				}

			} else {
				system.clearRun(explosionInterval);
			}
		}, hfData.hfSpeed);
	}


	if (hfData.hfAnim === 5) {
		let filledCount = 0;
		let level = 0; // Track the Y-levels to create the cascading effect

		const cascadeInterval = system.runInterval(() => {
			if (filledCount < airBlocksToFill.length) {
				for (const blockPos of airBlocksToFill) {
					if (blockPos.y === hfData.location.y + level) {
						const block = dimension.getBlock(blockPos);
						if (airLikeBlocks.includes(block.typeId)) {
							if (hfData.hfBlock === 0) {
								block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
							} else {
								block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
							}
							filledCount++;
						}
					}
				}

				level++;

				if (filledCount >= airBlocksToFill.length || level > 5) {
					system.clearRun(cascadeInterval);
					if (hfData.hfNotify === true) {
						player.sendMessage(`Filled ${filledCount} blocks.`);
					}
				}

			} else {
				system.clearRun(cascadeInterval);
			}
		}, hfData.hfSpeed + 1);
	}

	if (hfData.hfAnim === 6) {
		let filledCount = 0;
		const blocksPerTick = 10; // Number of blocks to place per tick
		const placedBlocks = new Set(); // Track filled blocks to prevent double filling
		let remainingBlocks = [...airBlocksToFill]; // Copy the array to avoid modifying the original

		const randomInterval = system.runInterval(() => {
			let blocksPlacedThisTick = 0;

			// Loop to place blocks within the limit for each tick
			while (blocksPlacedThisTick < blocksPerTick && remainingBlocks.length > 0) {
				const randomIndex = Math.floor(Math.random() * remainingBlocks.length);
				const blockPos = remainingBlocks[randomIndex];
				const blockKey = `${blockPos.x},${blockPos.y},${blockPos.z}`; // Unique key for each block

				// Check if the block has not already been filled
				if (!placedBlocks.has(blockKey)) {
					const block = dimension.getBlock(blockPos);
					if (airLikeBlocks.includes(block.typeId)) {
						if (hfData.hfBlock === 0) {
							block.setPermutation(BlockPermutation.resolve(hfData.blockData.typeId, hfData.blockData.states));
						} else {
							block.setPermutation(BlockPermutation.resolve(blockTypeIds[hfData.hfBlock]));
						}
						placedBlocks.add(blockKey); // Mark block as filled
						filledCount++;
						blocksPlacedThisTick++;
					}
				}

				// Remove the block from the array to prevent it from being selected again
				remainingBlocks.splice(randomIndex, 1);
			}

			// Stop when all blocks are filled
			if (remainingBlocks.length === 0) {
				system.clearRun(randomInterval);
				if (hfData.hfNotify === true) {
					player.sendMessage(`Filled ${filledCount} blocks.`);
				}
			}
		}, hfData.hfSpeed); // Run every tick for smoother block placement
	}
});


world.afterEvents.chatSend.subscribe((event) => {
	const { sender, message } = event;
	if (["?hf", "?dbgHoleFiller", "?HoleFiller", "?holefiller", "?HF", "?dbgholefiller"].includes(message)) {
		system.run(() => {
			const targetPlayerinv = sender.getComponent("inventory");
			const container = targetPlayerinv.container;
			const maxSlots = 36;

			let hasSettingsStick = false;
			let hasHoleFillerSnowballs = false;

			// Check if the player already has the custom stick or snowballs
			for (let i = 0; i < maxSlots; i++) {
				const item = container.getItem(i);
				if (!item) continue;
				if (item.typeId === "minecraft:stick" && item.nameTag === "Hole Filler Settings") {
					hasSettingsStick = true;
				}
				if (item.typeId === "minecraft:snowball" && item.nameTag === "Hole Filler") {
					hasHoleFillerSnowballs = true;
				}
			}
			// Track free slots
			const freeSlots = [];
			for (let i = 0; i < maxSlots && freeSlots.length < 2; i++) {
				const item = container.getItem(i);
				if (!item?.typeId) {
					freeSlots.push(i);
				}
			}
			if ((!hasHoleFillerSnowballs || !hasSettingsStick) && freeSlots.length === 0) {
				sender.sendMessage(`§cNot enough free slots! Free your inventory`);
				return;
			}
			if ((hasHoleFillerSnowballs || hasSettingsStick) && freeSlots.length === 0) {
				sender.sendMessage(`§cNot enough free slots! Free your inventory, §dit also appears that you already have the items`);
				return;
			}
			if ((!hasHoleFillerSnowballs && !hasSettingsStick) && freeSlots.length > 0) {
				sender.sendMessage("§gThanks for downloading Hole Filler Add-on, I hope you'll enjoy this!");
			}
			// Give snowballs if not already present
			if (freeSlots.length > 0) {
				const item1 = new ItemStack("minecraft:snowball", 16);
				item1.nameTag = "Hole Filler";
				container.setItem(freeSlots.shift(), item1);

				if (hasSettingsStick) {
					sender.sendMessage("§dMore Hole Filling Balls Given to you");
				}
			}
			// Give settings stick if not already present
			if (!hasSettingsStick && freeSlots.length > 0) {
				const item2 = new ItemStack("minecraft:stick", 1);
				item2.nameTag = "Hole Filler Settings";
				container.setItem(freeSlots.shift(), item2);
				if (hasHoleFillerSnowballs) {
					sender.sendMessage("§dMore Hole Filler Settings stick given to you[");
				}
			}
		});
		event.cancel = true;
		return;
	}
});